/*
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2003 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* includes */
#include "pmacct.h"
#include "plugin_hooks.h"
#include "imt_plugin.h"

/* global vars */
int got_hup_signal;

/* Functions */
void imt_plugin(int pipe_fd) 
{
  struct sockaddr cAddr;
  struct pkt_primitives addr;
  struct acc *acc_elem;
  struct pkt_data *data;
  unsigned char srvbuf[SRVBUFLEN];
  unsigned char pipebuf[SRVBUFLEN];
  char path[] = "/tmp/collect.pipe";
  unsigned int insert_status;

  fd_set read_descs; /* select() stuff */
  int select_fd;
  int cLen, num, sd, sd2;

  /* a bunch of default definitions and post-checks */
  no_more_space = FALSE;
  if ((!config.num_memory_pools) && (!have_num_memory_pools))
    config.num_memory_pools = NUM_MEMORY_POOLS;
  
  if (!config.memory_pool_size) config.memory_pool_size = MEMORY_POOL_SIZE;  
  else {
    if (config.memory_pool_size < sizeof(struct acc)) {
      printf("ERROR: invalid size of memory pools\n");
      exit(1);
    }
  }

  if (!config.imt_plugin_path) config.imt_plugin_path = path; 
  if (!config.buckets) config.buckets = MAX_HOSTS;

  init_memory_pool_table(config);
  if (mpd == NULL) {
    printf("ERROR: Cannot allocate memory pools table\n");
    exit(1);
  }

  current_pool = request_memory_pool(config.buckets*sizeof(struct acc));
  if (current_pool == NULL) {
    printf("ERROR: Cannot allocate my first memory pool, try with larger value.\n");
    exit(1);
  }

  a = init_accounting_structure();
  current_pool = request_memory_pool(config.memory_pool_size);
  if (current_pool == NULL) {
    printf("ERROR: Cannot allocate more memory pools, try with larger value.\n");
    exit(1);
  }

  signal(SIGHUP, my_sighup_handler); /* needed for clearing stats */

  /* building a server for interrogations by clients */
  sd = build_query_server(config.imt_plugin_path);
  cLen = sizeof(cAddr);

  /* preparing for synchronous I/O multiplexing */
  select_fd = 0;
  FD_ZERO(&read_descs);
  FD_SET(sd, &read_descs);
  if (sd > select_fd) select_fd = sd;
  FD_SET(pipe_fd, &read_descs);
  if (pipe_fd > select_fd) select_fd = pipe_fd;
  select_fd++;

  /* plugin main loop */
  while(1) {
    select(select_fd, &read_descs, NULL, NULL, NULL);

    /* doing server tasks */
    sd2 = accept(sd, &cAddr, &cLen);
    if (sd2 > 0) {
      switch (fork()) {
      case 0: /* Child */
        close(sd);
        num = recv(sd2, srvbuf, SRVBUFLEN, 0);
        if (num > 0) process_query_data(sd2, srvbuf, num);
        if (debug) printf("Closing connection ...\n");
        close(sd2);
        exit(1);
      default: /* Parent */
        close(sd2);
      }
    }

    /* clearing stats if necessary */
    if (got_hup_signal) {
      clear_memory_pool_descriptors();
      current_pool = request_memory_pool(config.buckets*sizeof(struct acc));
      if (current_pool == NULL) {
        printf("ERROR: Cannot allocate my first memory pool, try with larger value.\n");
        exit(1);
      }
      a = init_accounting_structure();
      current_pool = request_memory_pool(config.memory_pool_size);
      if (current_pool == NULL) {
        printf("ERROR: Cannot allocate more memory pools, try with larger value.\n");
        exit(1);
      }
      got_hup_signal = FALSE;
      no_more_space = FALSE;
    }

    if (read(pipe_fd, pipebuf, sizeof(struct pkt_data)) == 0)
      exit(0); /* we exit silently; something happened at the write end */

    if (debug) printf("**********\n");
    data = (struct pkt_data *) pipebuf;
    memset(&addr, 0, sizeof(struct pkt_primitives));
    insert_status = 0;
    acc_elem = NULL;

    /* 
       simple logic:
       while evaluating what to count, options are
       either stackable (src_*, dst_*) with boolean
       "and" or iterative (sum_*).
    */

    if (config.what_to_count & COUNT_SRC_MAC)
      memcpy(addr.eth_shost, data->primitives.eth_shost, ETH_ADDR_LEN);

    if (config.what_to_count & COUNT_DST_MAC)
      memcpy(addr.eth_dhost, data->primitives.eth_dhost, ETH_ADDR_LEN);

    if (config.what_to_count & COUNT_SRC_HOST)
      addr.ipa.s_addr = data->primitives.ipa.s_addr;

    if (config.what_to_count & COUNT_DST_HOST)
      addr.flow.s_addr = data->primitives.flow.s_addr;
    
    if (config.what_to_count & COUNT_SRC_PORT) {
      if ((data->primitives.proto == IPPROTO_UDP) || (data->primitives.proto == IPPROTO_TCP))
        addr.src_port = ntohs(data->primitives.src_port);
      else addr.src_port = 0;
    }

    if (config.what_to_count & COUNT_DST_PORT) { 
      if ((data->primitives.proto == IPPROTO_UDP) || (data->primitives.proto == IPPROTO_TCP))
        addr.dst_port = ntohs(data->primitives.dst_port);
      else addr.dst_port = 0;
    }

    if (config.what_to_count & COUNT_IP_PROTO) 
      addr.proto = data->primitives.proto;

    if (config.what_to_count & COUNT_SUM_HOST) {
      addr.ipa.s_addr = data->primitives.ipa.s_addr;
      acc_elem = insert_accounting_structure(&addr, ntohs(data->pkt_len));
      addr.ipa.s_addr = data->primitives.flow.s_addr;
      acc_elem = insert_accounting_structure(&addr, ntohs(data->pkt_len));
      insert_status |= INSERT_ALREADY_DONE;
    }

    if (!insert_status)
      acc_elem = insert_accounting_structure(&addr, ntohs(data->pkt_len));

    if (debug) debug_packet(data); 
  }
}
