/*
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2003 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#define __MYSQL_PLUGIN_C

/* includes */
#include "pmacct.h"
#include "pmacct-data.h"
#include "plugin_hooks.h"
#include "mysql_plugin.h"
#include "util.h"
#include "crc32.c"

/* Functions */
void mysql_plugin(int pipe_fd) 
{
  struct pkt_data empty_data, *data;
  struct pkt_primitives srcdst;
  unsigned char pipebuf[SRVBUFLEN];
  char mysql_user[] = "pmacct";
  char mysql_pwd[] = "arealsmartpwd";
  char mysql_db[] = "pmacct"; 
  char mysql_table[] = "acct";
  struct pollfd pfd;
  struct insert_data idata;
  time_t t, now, refresh_deadline;
  int timeout = config.sql_refresh_time*1000; /* dirty */ 
  int ret;

  /* signal handling */
  signal(SIGINT, exit_gracefully);
  if (config.daemon) signal(SIGHUP, exit_gracefully);
  
  memset(mysql_data, 0, sizeof(mysql_data));
  memset(cache, 0, sizeof(cache));
  memset(queries_queue, 0, sizeof(queries_queue));
  memset(&empty_data, 0, sizeof(struct pkt_data));
  memset(lock_clause, 0, sizeof(lock_clause));
  memset(unlock_clause, 0, sizeof(unlock_clause));
  memset(update_clause, 0, sizeof(update_clause));
  memset(insert_clause, 0, sizeof(insert_clause));
  memset(where, 0, sizeof(where));
  memset(values, 0, sizeof(values));

  qq_ptr = 0;
  pp_size = sizeof(struct pkt_primitives);
  pfd.fd = pipe_fd;
  pfd.events = POLLIN;
  setnonblocking(pipe_fd);

  /* searching for user supplied values */ 
  if (!config.sql_user) config.sql_user = mysql_user; 
  if (!config.sql_db) config.sql_db = mysql_db;
  if (!config.sql_passwd) config.sql_passwd = mysql_pwd; 
  if (!config.sql_table) config.sql_table = mysql_table;

  mysql_init(&db_desc); 
  if (mysql_real_connect(&db_desc, config.sql_host, config.sql_user, config.sql_passwd, config.sql_db, 0, NULL, 0) == NULL) {
    printf("%s\n", mysql_error(&db_desc));
    exit(1);
  }

  now = time(NULL);

  /* time handling */
  if (config.sql_history) {
    idata.basetime = now; 
    if (config.sql_history == COUNT_MINUTELY)
      idata.timeslot = config.sql_history_howmany*60;
    else if (config.sql_history == COUNT_HOURLY)
      idata.timeslot = config.sql_history_howmany*3600;
    else if (config.sql_history == COUNT_DAILY)
      idata.timeslot = config.sql_history_howmany*86400; 

    /* round off stuff */
    t = roundoff_time(idata.basetime);
    while ((t+idata.timeslot) < idata.basetime) t += idata.timeslot; 
    idata.basetime = t;
  }

  /* sql refresh deadline init */
  refresh_deadline = now; 
  t = roundoff_time(refresh_deadline);
  while ((t+config.sql_refresh_time) < refresh_deadline) t += config.sql_refresh_time;
  refresh_deadline = t;
  refresh_deadline += config.sql_refresh_time; /* it's a deadline not a basetime */

  /* setting number of entries in _protocols structure */
  while (_protocols[protocols_number].number != -1) protocols_number++;

  /* composing the proper (filled with primitives used during
     the current execution) SQL strings */
  idata.num_primitives = compose_static_queries();
  num_primitives = idata.num_primitives; 

  /* plugin main loop */
  while(1) {
    memset(&srcdst, 0, sizeof(struct pkt_primitives)); 
    memset(pipebuf, 0, sizeof(struct pkt_data));

    ret = poll(&pfd, 1, timeout);
    now = time(NULL);

    switch (ret) {
    case 0: /* timeout */
      if (qq_ptr) {
        switch (fork()) {
        case 0: /* Child */
          cache_purge(queries_queue, qq_ptr, idata.num_primitives);
          exit(0);
        default: /* Parent */
          qq_ptr = cache_flush(queries_queue, qq_ptr);
          break;
        }
      }
      break;
    default: /* we received data */
      if (read(pipe_fd, pipebuf, sizeof(struct pkt_data)) == 0) 
        exit(0); /* we exit silently; something happened at the write end */

      /* lazy sql refresh handling */ 
      if (now > refresh_deadline) {
        if (qq_ptr) {
          switch (fork()) {
          case 0: /* Child */
            cache_purge(queries_queue, qq_ptr, idata.num_primitives);
            exit(0);
          default: /* Parent */
            qq_ptr = cache_flush(queries_queue, qq_ptr);
	    refresh_deadline += config.sql_refresh_time; 
            break;
	  }
        }
      } 

      /* test: did we get something from read() ? */
      if (memcmp(pipebuf, &empty_data, sizeof(struct pkt_data))) {
        data = (struct pkt_data *) pipebuf;

	if (now > (idata.basetime + idata.timeslot)) idata.basetime += idata.timeslot;

        if (config.what_to_count & COUNT_SRC_MAC)
          memcpy(srcdst.eth_shost, data->primitives.eth_shost, ETH_ADDR_LEN);

        if (config.what_to_count & COUNT_DST_MAC)
          memcpy(srcdst.eth_dhost, data->primitives.eth_dhost, ETH_ADDR_LEN);

        if (config.what_to_count & COUNT_SRC_HOST)
          srcdst.ipa = data->primitives.ipa;

        if (config.what_to_count & COUNT_DST_HOST)
          srcdst.flow = data->primitives.flow;

        if (config.what_to_count & COUNT_SRC_PORT) {
          if ((data->primitives.proto == IPPROTO_UDP) || (data->primitives.proto == IPPROTO_TCP))
            srcdst.src_port = ntohs(data->primitives.src_port);
          else srcdst.src_port = 0;
        }

        if (config.what_to_count & COUNT_DST_PORT) {
          if ((data->primitives.proto == IPPROTO_UDP) || (data->primitives.proto == IPPROTO_TCP))
            srcdst.dst_port = ntohs(data->primitives.dst_port);
          else srcdst.dst_port = 0;
        }

	if (config.what_to_count & COUNT_SUM_HOST) {
	  printf("ERROR: Option available only in memory table operations\n");
	  printf("Exiting ...\n");
	  exit(0);
	}
	  
        if (config.what_to_count & COUNT_IP_PROTO) 
          srcdst.proto = data->primitives.proto;

	idata.modulo = cache_modulo(&srcdst); 
	cache_insert(&srcdst, data->pkt_len, &idata); 
      }
    }
  }
}

unsigned int cache_modulo(struct pkt_primitives *srcdst)
{
  register unsigned int modulo;

  modulo = cache_crc32((char *)srcdst, pp_size);
  
  return modulo %= CACHE_ENTRIES;
}

void cache_aliasing(const struct db_cache *cache_elem, const int num_primitives)
{
  if (debug) printf("*** Entry aliasing ***\n");
  if (mysql_query(&db_desc, lock_clause) != 0) {
    db_error_handling(&db_desc);
    printf("WARN: writing data to Mysql DB without locking ...\n");
  }

  cache_dbop(&db_desc, cache_elem, num_primitives);
  mysql_query(&db_desc, unlock_clause);
}

void cache_dbop(MYSQL *db_desc, const struct db_cache *cache_elem, const int num_primitives)
{
  char *ptr_values, *ptr_where;
  int num=0;

  ptr_where = where_clause;
  ptr_values = values_clause; 
  while (num < num_primitives) {
    (*where[num].handler)(cache_elem, num, &ptr_values, &ptr_where);
    num++;
  }
  
  /* constructing sql query */
  snprintf(mysql_data, sizeof(mysql_data), update_clause, cache_elem->packet_counter, cache_elem->bytes_counter);
  strncat(mysql_data, where_clause, SPACELEFT(mysql_data));
  if (mysql_query(db_desc, mysql_data) != 0) db_error_handling(db_desc);

  if (mysql_affected_rows(db_desc) == 0) {
    /* UPDATE failed, trying with an INSERT query */ 
    strncpy(mysql_data, insert_clause, sizeof(mysql_data));
    snprintf(ptr_values, SPACELEFT(values_clause), ", %d, %d)", cache_elem->packet_counter, cache_elem->bytes_counter);
    strncat(mysql_data, values_clause, SPACELEFT(mysql_data));
    if (mysql_query(db_desc, mysql_data) != 0) db_error_handling(db_desc);
  }

  if (debug) {
    printf("**********\n");
    printf("%s\n", mysql_data);
  }
}

void cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct insert_data *idata) 
{
  if (memcmp(&cache[idata->modulo], srcdst, sizeof(struct pkt_primitives)) != 0) { 
    /* aliasing of entries: if there is some valid value,
       we trap a live sql query; otherwise we put this entry
       in the queue of pending sql queries */
    if (cache[idata->modulo].valid == TRUE) cache_aliasing(&cache[idata->modulo], idata->num_primitives); 
    else {
      queries_queue[qq_ptr] = &cache[idata->modulo];
      qq_ptr++;
    }

    /* we add the new entry in the cache */
    memcpy(&cache[idata->modulo], srcdst, sizeof(struct pkt_primitives));
    cache[idata->modulo].packet_counter = 1;
    cache[idata->modulo].bytes_counter = ntohs(ip_len);
    cache[idata->modulo].valid = TRUE;
    cache[idata->modulo].basetime = idata->basetime;
  }
  else {
    if (cache[idata->modulo].valid == TRUE) {
      /* additional check: time */
      if ((cache[idata->modulo].basetime < idata->basetime) && (config.sql_history)) {
	cache_aliasing(&cache[idata->modulo], idata->num_primitives);
	cache[idata->modulo].packet_counter = 1;
	cache[idata->modulo].bytes_counter = ntohs(ip_len);
	cache[idata->modulo].basetime = idata->basetime;
      }
      else {
        /* everything is ok; summing counters */
        cache[idata->modulo].packet_counter++;
        cache[idata->modulo].bytes_counter += ntohs(ip_len);
      }
    }
    else {
      /* entry invalidated by SIGALRM; restarting counters */
      cache[idata->modulo].packet_counter = 1;
      cache[idata->modulo].bytes_counter = ntohs(ip_len);
      cache[idata->modulo].valid = TRUE;
      cache[idata->modulo].basetime = idata->basetime;
      queries_queue[qq_ptr] = &cache[idata->modulo];
      qq_ptr++;
    }
  }
}

void cache_purge(struct db_cache *queue[], int index, const int num_primitives)
{
  int j;
  MYSQL db_desc;

  /* we have to ignore signals to avoid loops:
     because we are already fork()ed */ 
  signal(SIGINT, SIG_IGN);
  signal(SIGHUP, SIG_IGN);

  /* we are invoking this function after a fork();
     so it's needed brand new connections with the
     DB */
  mysql_init(&db_desc);
  if (mysql_real_connect(&db_desc, config.sql_host, config.sql_user, config.sql_passwd, config.sql_db, 0, NULL, 0) == NULL) {
    printf("%s\n", mysql_error(&db_desc));
    exit(1);
  }

  if (debug) printf("*** Purging cache - START ***\n");
  if (mysql_query(&db_desc, lock_clause) != 0) {
    db_error_handling(&db_desc);
    printf("WARN: writing data to Mysql DB without locking ...\n");
  }

  for (j = 0; j < index; j++) cache_dbop(&db_desc, queue[j], num_primitives);

  /* rewinding stuff */
  mysql_query(&db_desc, unlock_clause);
  mysql_close(&db_desc);
  if (debug) printf("*** Purging cache - END ***\n");
}

int cache_flush(struct db_cache *queue[], int index)
{
  int j;

  for (j = 0; j < index; j++) queue[j]->valid = FALSE;
  index = 0;
  
  return index;
}

void db_error_handling(MYSQL *db_desc)
{
  /* an error occurred: taking countermeasures */
  printf("Error (msg): %s\n", mysql_error(db_desc));
}

int evaluate_history(int primitive)
{
  char timestamp_format_string[SRVBUFLEN];
  
  if (config.sql_history) {
    strcpy(timestamp_format_string, "\'%%Y-%%m-%%d %%H:%%i\'");

    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(where[primitive].string, "FROM_UNIXTIME(%d, ", SPACELEFT(where[primitive].string));
    strncat(where[primitive].string, timestamp_format_string, SPACELEFT(where[primitive].string));
    strncat(where[primitive].string, ") = DATE_FORMAT(stamp_inserted, ", SPACELEFT(where[primitive].string));
    strncat(where[primitive].string, timestamp_format_string, SPACELEFT(where[primitive].string));
    strncat(where[primitive].string, ")", SPACELEFT(where[primitive].string));

    strncat(insert_clause, "stamp_updated, stamp_inserted", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "now(), FROM_UNIXTIME(%d)", SPACELEFT(values[primitive].string));

    where[primitive].type = values[primitive].type = TIMESTAMP;
    values[primitive].handler = where[primitive].handler = count_timestamp_handler;
    primitive++;
  }

  return primitive;
}

int evaluate_primitives(int primitive)
{
  register unsigned long int what_to_count=0;

  if (config.sql_optimize_clauses) what_to_count = config.what_to_count;
  else {
    /* we are requested to avoid optimization;
       then we'll construct an all-true "what
       to count" bitmap */ 
    what_to_count |= COUNT_SRC_HOST|COUNT_DST_HOST;
    what_to_count |= COUNT_SRC_MAC|COUNT_DST_MAC;
    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO;
  }

  /* 1st part: arranging pointers to an opaque structure and 
     composing the static selection (WHERE) string */

  if (what_to_count & COUNT_SRC_MAC) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "mac_src", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%s\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "mac_src=\'%s\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_SRC_MAC;
    values[primitive].handler = where[primitive].handler = count_src_mac_handler;
    primitive++;
  }

  if (what_to_count & COUNT_DST_MAC) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "mac_dst", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%s\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "mac_dst=\'%s\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_DST_MAC;
    values[primitive].handler = where[primitive].handler = count_dst_mac_handler;
    primitive++;
  }

  if (what_to_count & COUNT_SRC_HOST) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "ip_src", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%s\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "ip_src=\'%s\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_SRC_HOST;
    values[primitive].handler = where[primitive].handler = count_src_host_handler;
    primitive++;
  }

  if (what_to_count & COUNT_DST_HOST) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "ip_dst", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%s\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "ip_dst=\'%s\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_DST_HOST;
    values[primitive].handler = where[primitive].handler = count_dst_host_handler;
    primitive++;
  }

  if (what_to_count & COUNT_SRC_PORT) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "src_port", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%d\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "src_port=\'%d\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_SRC_PORT;
    values[primitive].handler = where[primitive].handler = count_src_port_handler;
    primitive++;
  }

  if (what_to_count & COUNT_DST_PORT) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "dst_port", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%d\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "dst_port=\'%d\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_DST_PORT;
    values[primitive].handler = where[primitive].handler = count_dst_port_handler;
    primitive++;
  }

  if (what_to_count & COUNT_IP_PROTO) {
    if (primitive) {
      strncat(insert_clause, ", ", SPACELEFT(insert_clause));
      strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
      strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
    }
    strncat(insert_clause, "ip_proto", SPACELEFT(insert_clause));
    strncat(values[primitive].string, "\'%s\'", SPACELEFT(values[primitive].string));
    strncat(where[primitive].string, "ip_proto=\'%s\'", SPACELEFT(where[primitive].string));
    values[primitive].type = where[primitive].type = COUNT_IP_PROTO;
    values[primitive].handler = where[primitive].handler = count_ip_proto_handler;
    primitive++;
  }

  return primitive;
}

int compose_static_queries()
{
  int primitives=0;

  /* "INSERT INTO ... VALUES ... " and "... WHERE ..." stuff */
  strncpy(where[primitives].string, " WHERE ", sizeof(where[primitives].string));
  snprintf(insert_clause, sizeof(insert_clause), "INSERT INTO %s (", config.sql_table);
  strncpy(values[primitives].string, " VALUES (", sizeof(values[primitives].string));
  primitives = evaluate_history(primitives);
  primitives = evaluate_primitives(primitives);
  strncat(insert_clause, ", packets, bytes)", SPACELEFT(insert_clause));

  /* "LOCK ..." stuff */
  snprintf(lock_clause, sizeof(lock_clause), "LOCK TABLES %s WRITE", config.sql_table);
  strncpy(unlock_clause, "UNLOCK TABLES", sizeof(unlock_clause));

  /* "UPDATE ... SET ..." stuff */
  snprintf(update_clause, sizeof(update_clause), "UPDATE %s ", config.sql_table);
  strncat(update_clause, "SET packets=packets+%d, bytes=bytes+%d", SPACELEFT(update_clause));
  if (config.sql_history) strncat(update_clause, ", stamp_updated=now()", SPACELEFT(update_clause));

  return primitives;
}

void exit_gracefully(int signum)
{
  printf("*** Purging MySQL queries queue ***\n");
  cache_purge(queries_queue, qq_ptr, num_primitives);
  exit(0); 
}

