/*  
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2003 by Paolo Lucente
*/

#define min(a,b) ((a)>(b)?(b):(a))
#define EXTRACT_16BITS(p) \
  ((u_int16_t)*((const u_int8_t *)(p) + 0) << 8 | \
  (u_int16_t)*((const u_int8_t *)(p) + 1))

#define ETH_ADDR_LEN    	6               /* Octets in one ethernet addr   */
#define ETHER_HDRLEN    	14
#define ETHERMTU		1500
#define IEEE8021Q_TAGLEN	4

/* 10Mb/s ethernet header */
struct eth_header
{
  u_int8_t  ether_dhost[ETH_ADDR_LEN];      /* destination eth addr */
  u_int8_t  ether_shost[ETH_ADDR_LEN];      /* source ether addr    */
  u_int16_t ether_type;                     /* packet type ID field */
};

/* Ethernet protocol ID's */
#define ETHERTYPE_IP		0x0800          /* IP */
#define ETHERTYPE_8021Q		0x8100          /* 802.1Q */

struct fddi_header {
  u_char  fddi_fc;                /* frame control */
  u_char  fddi_dhost[6];
  u_char  fddi_shost[6];
};

#define FDDI_HDRLEN 13

/* Useful values for fddi_fc (frame control) field */

/*
 * FDDI Frame Control bits
 */
#define FDDIFC_C                0x80            /* Class bit */
#define FDDIFC_L                0x40            /* Address length bit */
#define FDDIFC_F                0x30            /* Frame format bits */
#define FDDIFC_Z                0x0f            /* Control bits */

/*
 * FDDI Frame Control values. (48-bit addressing only).
 */
#define FDDIFC_VOID             0x40            /* Void frame */
#define FDDIFC_NRT              0x80            /* Nonrestricted token */
#define FDDIFC_RT               0xc0            /* Restricted token */
#define FDDIFC_SMT_INFO         0x41            /* SMT Info */
#define FDDIFC_SMT_NSA          0x4F            /* SMT Next station adrs */
#define FDDIFC_MAC_BEACON       0xc2            /* MAC Beacon frame */
#define FDDIFC_MAC_CLAIM        0xc3            /* MAC Claim frame */
#define FDDIFC_LLC_ASYNC        0x50            /* Async. LLC frame */
#define FDDIFC_LLC_SYNC         0xd0            /* Sync. LLC frame */
#define FDDIFC_IMP_ASYNC        0x60            /* Implementor Async. */
#define FDDIFC_IMP_SYNC         0xe0            /* Implementor Synch. */
#define FDDIFC_SMT              0x40            /* SMT frame */
#define FDDIFC_MAC              0xc0            /* MAC frame */

#define FDDIFC_CLFF             0xF0            /* Class/Length/Format bits */
#define FDDIFC_ZZZZ             0x0F            /* Control bits */

/*
 * A somewhat abstracted view of the LLC header
 */

struct llc {
        u_int8_t dsap;
        u_int8_t ssap;
        union {
                u_int8_t u_ctl;
                u_int16_t is_ctl;
                struct {
                        u_int8_t snap_ui;
                        u_int8_t snap_pi[5];
                } snap;
                struct {
                        u_int8_t snap_ui;
                        u_int8_t snap_orgcode[3];
                        u_int8_t snap_ethertype[2];
                } snap_ether;
        } ctl;
};

/*
 * LLC/SNAP
 */
#define LLC_GSAP        1
#define LLC_S_FMT       1
#define LLC_U_POLL      0x10
#define LLC_IS_POLL     0x0100
#define LLC_XID_FI      0x81
#define LLC_UI          0x03
#define LLC_UA          0x63
#define LLC_DISC        0x43
#define LLC_DM          0x0f
#define LLC_SABME       0x6f
#define LLC_TEST        0xe3
#define LLC_XID         0xaf
#define LLC_FRMR        0x87
#define LLC_RR          0x0001
#define LLC_RNR         0x0005
#define LLC_REJ         0x0009

#define LLCSAP_NULL             0x00
#define LLCSAP_IP               0x06
#define LLCSAP_SNAP             0xaa

/* PPP protocol definitions */
#define PPP_HDRLEN      4       /* octets for standard ppp header */
#define PPP_IP          0x21    /* Internet Protocol */

/*
 * Copyright (c) 1982, 1986, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)ip.h        8.1 (Berkeley) 6/10/93
 */

/*
 * Definitions for internet protocol version 4.
 * Per RFC 791, September 1981.
 */

/*
 * Structure of an internet header, naked of options.
 */

struct my_iphdr
{
   u_int8_t     ip_vhl;         /* header length, version */
#define IP_V(ip)        (((ip)->ip_vhl & 0xf0) >> 4)
#define IP_HL(ip)       ((ip)->ip_vhl & 0x0f)
   u_int8_t     ip_tos;         /* type of service */
   u_int16_t    ip_len;         /* total length */
   u_int16_t    ip_id;          /* identification */
   u_int16_t    ip_off;         /* fragment offset field */
#define IP_DF 0x4000                    /* dont fragment flag */
#define IP_MF 0x2000                    /* more fragments flag */
#define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */
   u_int8_t     ip_ttl;         /* time to live */
   u_int8_t     ip_p;           /* protocol */
   u_int16_t    ip_sum;         /* checksum */
   struct in_addr ip_src;	/* source and destination addresses */
   struct in_addr ip_dst; 
};

struct my_tlhdr {
   u_int16_t	src_port;	/* source and destination ports */
   u_int16_t	dst_port;
};

struct packet_ptrs {
  u_char *packet_ptr; /* ptr to the whole packet */
  u_char *mac_ptr; /* ptr to mac addresses */
  u_char *iph_ptr; /* ptr to ip header */
  u_char *tlh_ptr; /* ptr to transport level protocol header */
};

struct pkt_primitives {
  u_int8_t eth_dhost[ETH_ADDR_LEN];
  u_int8_t eth_shost[ETH_ADDR_LEN];
  struct in_addr ipa;
  struct in_addr flow;
  u_int16_t src_port;
  u_int16_t dst_port;
  u_int8_t proto;
};

struct pkt_data {
  struct pkt_primitives primitives;
  int pkt_len;
};
