/*  
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2003 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* defines */
#define __PMACCTD_C

/* includes */
#include "pmacct.h"
#include "pmacct-data.h"
#include "plugin_hooks.h"

/* variables to be exported away */
int debug, got_alrm_signal;
struct configuration config; /* plugin stuff */
int have_num_memory_pools; /* global getopt() stuff */
extern struct _configuration_line _configuration[];

/* Functions */
void usage_daemon(char *prog_name)
{
  printf("%s\n", COLLECTOR_USAGE_HEADER);
  printf("Usage: %s [-D] [-b buckets] [-i interface] [filter]\n", prog_name);
  printf("       %s [-f config_file]\n", prog_name);
  printf("       %s [-h]\n", prog_name);
  printf("\n");
  printf("Options:\n");
  printf("  -f  \tspecify a configuration file (see EXAMPLES file for valid config keys)\n");
  printf("  -c  \t[src_mac|dst_mac|src_host|dst_host|sum|src_port|dst_port|proto] \n\tcounts source, destination or total IP traffic (default src_host)\n");
  printf("  -D  \tdaemonize the process\n"); 
  printf("  -P  \t[mysql|memory] \n\tactivate specified plugins\n"); 
  printf("  -p  \tpath for client/server fifo (used by in-memory table plugin only)\n");
  printf("  -d  \tenables debug mode\n");
  printf("  -b  \tnumber of buckets\n");
  printf("  -i  \tinterface used for listening\n");
  printf("  -m  \tnumbers of memory pools (see INTERNALS file)\n");
  printf("  -r  \trefresh time of data into SQL database from in-memory cache (in seconds)\n");
  printf("  -s  \tsize of each memory pool\n");
  printf("  -h  \tprints this page\n");
  printf("\n");
  printf("Examples:\n");
  printf("  Daemonize the process; listen on eth0; write flow statistics in a mysql database\n"); 
  printf("  pmacctd -c dst_port -i eth0 -D -P mysql\n\n");
  printf("  Debug mode; listen on ee1; collect stats aggregating for destinations in a in-memory\n");
  printf("  structure with 8 memory pools of 64Kb each\n");
  printf("  pmacctd -c dst_host -i ee1 -d -m 8 -s 65535\n");
  printf("\n");
  printf("  See EXAMPLES file in the distribution for more examples\n");
  printf("\n");
  printf("For suggestions, critics, bugs, contact me: %s.\n", MANTAINER);
}


int main(int argc,char **argv)
{
  bpf_u_int32 localnet, netmask;  /* pcap library stuff */
  struct bpf_program filter;
  struct pcap_device device;
  char errbuf[PCAP_ERRBUF_SIZE];
  u_char *buf;
  u_short promisc = TRUE;
  int index;

  char count[128], *count_token, *count_ptr;
  char path[128], config_file[256];
  char plugins[128];
  int psize = DEFAULT_SNAPLEN;

  /* getopt() stuff */
  extern char *optarg;
  extern int optind, opterr, optopt;
  int errflag, cp; 

  /* select() stuff */
  fd_set read_descs; 
  int pcap_fd, select_fd;

  if (getuid() != 0) { 
    printf("You need superuser privileges to run this program. Exiting ...\n\n");
    exit(0);
  }

  /* a bunch of default definitions */ 
  strcpy(path, "/tmp/collect.pipe");
  debug = FALSE;
  have_num_memory_pools = FALSE;
  errflag = 0;

  use_inmemory_table = TRUE; /* default plugin configuration */
  memset(&config, 0, sizeof(struct configuration));
  memset(&device, 0, sizeof(struct pcap_device));
  memset(&config_file, 0, sizeof(config_file));

  /* getting commandline values */
  while (!errflag && ((cp = getopt(argc, argv, ARGS_DAEMON)) != -1)) {
    switch (cp) {
    case 'P':
      strlcpy(plugins, optarg, sizeof(plugins));
      evaluate_plugins(plugins);
      break;
    case 'D':
      config.daemon = TRUE;
      break;
    case 'd':
      debug = TRUE;
      break;
    case 'f':
      strlcpy(config_file, optarg, sizeof(config_file));
      config.using_config_file = TRUE;
      break;
    case 'c':
      strlcpy(count, optarg, sizeof(count));
      count_ptr = count;
      while (*count_ptr != '\0') { 
	count_token = extract_token(&count_ptr, ','); 
        if (!strcmp(count_token, "src_host")) config.what_to_count |= COUNT_SRC_HOST;
        else if (!strcmp(count_token, "dst_host")) config.what_to_count |= COUNT_DST_HOST;
        else if (!strcmp(count_token, "sum")) config.what_to_count |= COUNT_SUM_HOST;
        else if (!strcmp(count_token, "src_port")) config.what_to_count |= COUNT_SRC_PORT;
        else if (!strcmp(count_token, "dst_port")) config.what_to_count |= COUNT_DST_PORT;
        else if (!strcmp(count_token, "proto")) config.what_to_count |= COUNT_IP_PROTO;
	else if (!strcmp(count_token, "src_mac")) config.what_to_count |= COUNT_SRC_MAC;
	else if (!strcmp(count_token, "dst_mac")) config.what_to_count |= COUNT_DST_MAC;
        else printf("WARN: ignoring unknown aggregation method: %s.\n", count_token); 
      }
      break;
    case 'b':
      config.buckets = atoi(optarg);
      if (config.buckets <= 0) {
	printf("ERROR: invalid number of buckets\n");
	exit(1);
      }
      break;
    case 'm':
      config.num_memory_pools = atoi(optarg);
      if (config.num_memory_pools < 0) {
	printf("ERROR: invalid number of memory pools\n");
	exit(1);
      }
      have_num_memory_pools = TRUE;
      break;
    case 'p':
      strlcpy(path, optarg, sizeof(path));
      config.imt_plugin_path = path;
      break;
    case 'r':
      config.sql_refresh_time = atoi(optarg);
      if (config.sql_refresh_time <= 0) {
        printf("ERROR: invalid refresh time\n");
	exit(1);
      }
      break;
    case 's':
      config.memory_pool_size = atoi(optarg);
      break;
    case 'i':
      config.dev = optarg;
      break;
    case 'h':
      usage_daemon(argv[0]);
      exit(1);
      break;
    default:
      printf("ERROR: parameter %c unknown! \n  Exiting...\n\n", cp);
      usage_daemon(argv[0]);
      exit(1);
      break;
    }
  }

  /* post-checks and resolving conflicts */
  if (config.using_config_file) {
    if (parse_configuration_file(config_file) != SUCCESS) 
      exit(1);
    if (evaluate_configuration() != SUCCESS) 
      exit(1);
  }

  if (config.what_to_count & COUNT_SUM_HOST) {
    if (config.what_to_count != COUNT_SUM_HOST) {
      printf("ERROR: invalid combination of aggregation methods\n");
      exit(0);
    }
  }
  
  if (config.what_to_count == 0) {
    printf("WARN: defaulting to source ip aggregation method\n");
    config.what_to_count |= COUNT_SRC_HOST;
  }

  if (!config.dev) {
    if (debug) printf("WARN: Choosing a suitable device.\n");
    config.dev = pcap_lookupdev(errbuf); 
  }

  if (config.daemon) {
    if (!debug) {
      daemonize();
      signal(SIGSTOP, SIG_IGN);
      signal(SIGINT, SIG_IGN);
    }
    else printf("WARN: debug enabled; not fork()ing in background.\n");
  }

  if (!config.sql_refresh_time)
    config.sql_refresh_time = DEFAULT_DB_REFRESH_TIME;

  /* reading filter; if it exists, we'll take an action later */
  if (!config.using_config_file) 
    config.clbuf = copy_argv(&argv[optind]);

  signal(SIGCHLD, handle_falling_child); /* this eliminates zombies */
  signal(SIGHUP, SIG_IGN); /* we want to ignore hangup requests */
  signal(SIGPIPE, SIG_IGN); /* we want to exit gracefully when a pipe is broken */

  if ((device.dev_desc = pcap_open_live(config.dev, psize, promisc, 1000, errbuf)) == NULL) {
    printf("ERROR: pcap_open_live(): %s\n", errbuf);
    exit(1);
  } 
  glob_pcapt = device.dev_desc; /* SIGINT/stats handling */ 

  device.link_type = pcap_datalink(device.dev_desc); 
  for (index = 0; _devices[index].link_type != -1; index++) {
    if (device.link_type == _devices[index].link_type)
      device.data = &_devices[index];
  }

  if (device.data == NULL) {
    printf("ERROR: not supported data link type: %d\n", device.link_type); 
    exit(0);
  }
  
  /* doing pcap stuff;
     compiling filter even if it doesn't exists: libpcap 0.4 legacy */
  if (pcap_lookupnet(config.dev, &localnet, &netmask, errbuf) < 0) {
    localnet = 0;
    netmask = 0;
    printf("WARN: %s\n", errbuf);
  }

  if (pcap_compile(device.dev_desc, &filter, config.clbuf, 0, netmask) < 0)
    printf("ERROR: %s\nWARN: going on without a filter\n", pcap_geterr(device.dev_desc));

  if (pcap_setfilter(device.dev_desc, &filter) < 0)
    printf("ERROR: %s\nWARN: going on without a filter\n", pcap_geterr(device.dev_desc));

  /* plugins glue: creation */
  load_plugins(&config);

  /* signals to be handled only by pmacctd;
     we set proper handlers after plugin creation */
  if (!config.daemon) signal(SIGINT, my_sigint_handler);
  
  /* Main loop */
  while(1) pcap_dispatch(device.dev_desc, -1, pcap_cb, (u_char *) &device);
}

void pcap_cb(u_char *user, const struct pcap_pkthdr *pkthdr, const u_char *buf)
{
  struct packet_ptrs pptrs;
  struct pcap_device *device = (struct pcap_device *) user;

  /* We process the packet with the appropriate
     data link layer function */
  if (buf) { 
    pptrs.packet_ptr = (u_char *) buf;
    (*device->data->handler)(pkthdr, &pptrs);
    if (pptrs.iph_ptr) {
      ip_handler(&pptrs);
      exec_plugins(&pptrs);
    }
  }
} 

void ip_handler(register struct packet_ptrs *pptrs)
{
  register u_int8_t len = 0;
  struct my_iphdr *ip = (struct my_iphdr *) pptrs->iph_ptr;
  
  /* len: number of 32bit words forming the header */
  len = IP_HL(((struct my_iphdr *) pptrs->iph_ptr));
  pptrs->tlh_ptr = pptrs->iph_ptr+(len*4); 
}
